'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var all = _interopDefault(require('./configs/all.json'));
var base = _interopDefault(require('./configs/base.json'));
var eslintRecommended = _interopDefault(require('./configs/eslint-recommended.json'));
var functionalLite = _interopDefault(require('./configs/functional-lite.json'));
var functional = _interopDefault(require('./configs/functional.json'));
var recommended = _interopDefault(require('./configs/recommended.json'));
var experimentalUtils = require('@typescript-eslint/experimental-utils');
var escapeRegExp = _interopDefault(require('escape-string-regexp'));
var deepmerge = require('deepmerge');
var astNodeTypes = require('@typescript-eslint/typescript-estree/dist/ts-estree/ast-node-types');

const version = "0.0.0";

/**
 * This file has functions that typeguard the given node/type.
 */
function isCallExpression(node) {
    return node.type === "CallExpression";
}
function isClassLike(node) {
    return node.type === "ClassDeclaration" || node.type === "ClassExpression";
}
function isExpressionStatement(node) {
    return node.type === "ExpressionStatement";
}
function isForXInitialiser(node) {
    const forX = getForXStatement(node);
    return forX !== null && forX.left === node;
}
function isForXStatement(node) {
    return node.type === "ForInStatement" || node.type === "ForOfStatement";
}
function isFunctionLike(node) {
    return (node.type === "FunctionDeclaration" ||
        node.type === "FunctionExpression" ||
        node.type === "ArrowFunctionExpression");
}
function isIdentifier(node) {
    return node.type === "Identifier";
}
function isMemberExpression(node) {
    return node.type === "MemberExpression";
}
function isTSIndexSignature(node) {
    return node.type === "TSIndexSignature";
}
function isTSInterfaceBody(node) {
    return node.type === "TSInterfaceBody";
}
function isTSArrayType(node) {
    return node.type === "TSArrayType";
}
function isTSPropertySignature(node) {
    return node.type === "TSPropertySignature";
}
function isTSTypeOperator(node) {
    return node.type === "TSTypeOperator";
}
function isTypeAliasDeclaration(node) {
    return node.type === "TSTypeAliasDeclaration";
}
function isVariableDeclaration(node) {
    return node.type === "VariableDeclaration";
}
function isVariableDeclarator(node) {
    return node.type === "VariableDeclarator";
}

/**
 * Test if the given node is in a function.
 */
function inFunction(node) {
    return getParentOfType(isFunctionLike, node) !== null;
}
/**
 * Test if the given node is in a class.
 */
function inClass(node) {
    return getParentOfType(isClassLike, node) !== null;
}
/**
 * Test if the given node is in a TS Property Signature.
 */
function inInterface(node) {
    return getParentOfType(isTSInterfaceBody, node) !== null;
}
/**
 * Test if the given node is in a ForX Statememt.
 */
function getForXStatement(node) {
    return getParentOfType(isForXStatement, node);
}
/**
 * Return the parent that meets the given check criteria.
 */
function getParentOfType(checker, node) {
    let n = node;
    while ((n = n.parent)) {
        if (checker(n)) {
            return n;
        }
    }
    return null;
}

const ignoreLocalOptionSchema = {
    type: "object",
    properties: {
        ignoreLocal: {
            type: "boolean"
        }
    },
    additionalProperties: false
};
const ignorePatternOptionsSchema = {
    type: "object",
    properties: {
        ignorePattern: {
            type: ["string", "array"],
            items: {
                type: "string"
            }
        },
        ignorePrefix: {
            type: ["string", "array"],
            items: {
                type: "string"
            }
        },
        ignoreSuffix: {
            type: ["string", "array"],
            items: {
                type: "string"
            }
        }
    },
    additionalProperties: false
};
const ignoreReturnTypeOptionSchema = {
    type: "object",
    properties: {
        ignoreReturnType: {
            type: "boolean"
        }
    },
    additionalProperties: false
};
const ignoreClassOptionSchema = {
    type: "object",
    properties: {
        ignoreClass: {
            type: "boolean"
        }
    },
    additionalProperties: false
};
const ignoreInterfaceOptionSchema = {
    type: "object",
    properties: {
        ignoreInterface: {
            type: "boolean"
        }
    },
    additionalProperties: false
};
/**
 * Should the given node be ignored?
 */
function shouldIgnore(node, context, ignoreOptions) {
    // Ignore if in a function and ignore-local is set.
    if (ignoreOptions.ignoreLocal && inFunction(node)) {
        return true;
    }
    // Ignore if in a class and ignore-class is set.
    if (ignoreOptions.ignoreClass && inClass(node)) {
        return true;
    }
    // Ignore if in an interface and ignore-interface is set.
    if (ignoreOptions.ignoreInterface && inInterface(node)) {
        return true;
    }
    const identifiers = (isVariableDeclaration(node)
        ? node.declarations.map(declaration => isIdentifier(declaration.id) ? declaration.id.name : undefined)
        : isVariableDeclarator(node) || isTypeAliasDeclaration(node)
            ? isIdentifier(node.id)
                ? [node.id.name]
                : []
            : isExpressionStatement(node)
                ? isCallExpression(node.expression)
                    ? [context.getSourceCode().getText(node.expression.callee)]
                    : []
                : isTSPropertySignature(node)
                    ? isIdentifier(node.key)
                        ? [node.key.name]
                        : []
                    : []).filter(name => name !== undefined);
    if (identifiers.length > 0 &&
        identifiers.every(identifier => {
            // Ignore if ignore-pattern is set and the pattern matches.
            if (ignoreOptions.ignorePattern &&
                isIgnoredPattern(identifier, ignoreOptions.ignorePattern)) {
                return true;
            }
            // Ignore if ignore-prefix is set and the prefix matches.
            if (ignoreOptions.ignorePrefix &&
                isIgnoredPrefix(identifier, ignoreOptions.ignorePrefix)) {
                return true;
            }
            // Ignore if ignore-suffix is set and the suffix matches.
            if (ignoreOptions.ignoreSuffix &&
                isIgnoredSuffix(identifier, ignoreOptions.ignoreSuffix)) {
                return true;
            }
            return false;
        })) {
        return true;
    }
    return false;
}
function isIgnoredPrefix(text, ignorePrefix) {
    if (Array.isArray(ignorePrefix)) {
        if (ignorePrefix.find(pfx => text.indexOf(pfx) === 0)) {
            return true;
        }
    }
    else {
        if (text.indexOf(ignorePrefix) === 0) {
            return true;
        }
    }
    return false;
}
function isIgnoredSuffix(text, ignoreSuffix) {
    if (Array.isArray(ignoreSuffix)) {
        if (ignoreSuffix.find(sfx => {
            const indexToFindAt = text.length - sfx.length;
            return indexToFindAt >= 0 && text.indexOf(sfx) === indexToFindAt;
        })) {
            return true;
        }
    }
    else {
        const indexToFindAt = text.length - ignoreSuffix.length;
        if (indexToFindAt >= 0 && text.indexOf(ignoreSuffix) === indexToFindAt) {
            return true;
        }
    }
    return false;
}
function isIgnoredPattern(text, ignorePattern) {
    const patterns = Array.isArray(ignorePattern)
        ? ignorePattern
        : [ignorePattern];
    // One or more patterns match?
    return patterns.some(pattern => findMatch(pattern.split("."), text.split(".")));
}
function findMatch(patternParts, textParts) {
    let index = 0;
    for (; index < patternParts.length; index++) {
        // Out of text?
        if (index >= textParts.length) {
            return false;
        }
        switch (patternParts[index]) {
            // Match any depth (including 0)?
            case "**": {
                const subpattern = patternParts.slice(index + 1);
                for (let offset = index; offset < textParts.length; offset++) {
                    const submatch = findMatch(subpattern, textParts.slice(offset));
                    if (submatch) {
                        return submatch;
                    }
                }
                return false;
            }
            // Match anything?
            case "*":
                continue;
            default:
                break;
        }
        // textParts[i] matches patternParts[i]?
        if (new RegExp("^" + escapeRegExp(patternParts[index]).replace(/\\\*/g, ".*") + "$").test(textParts[index])) {
            continue;
        }
        // No Match.
        return false;
    }
    // Match.
    return textParts.length === index;
}

/**
 * Create a rule.
 */
function createRule(data) {
    return experimentalUtils.ESLintUtils.RuleCreator(name => `https://github.com/jonaskello/tslint-immutable/blob/v${version}/README.md#${name}`)(data);
}
/**
 * Create a function that processes common options and then runs the given
 * check.
 */
function checkNode(check, context, ignoreOptions, otherOptions = []) {
    return (node) => {
        if (ignoreOptions && shouldIgnore(node, context, ignoreOptions)) {
            return;
        }
        const options = [ignoreOptions, ...otherOptions].filter(option => option !== undefined);
        return check(node, context, options);
    };
}

// The name of this rule.
const name = "no-array-mutation";
// The default options for the rule.
const defaultOptions = [];
// The possible error messages.
const errorMessages = {
    generic: "Mutating an array is not allowed."
};
// The meta data for this rule.
const meta = {
    type: "suggestion",
    docs: {
        description: "Disallow mutating arrays.",
        category: "Best Practices",
        recommended: "error"
    },
    messages: errorMessages,
    schema: []
};
/**
 * Check if the given node violates this rule.
 */
function check(node, context) {
    // TODO: port rule.
    context.report({ node, messageId: "generic" });
}
// Create the rule.
const rule = createRule({
    name,
    meta,
    defaultOptions,
    create(context, options) {
        const _checkNode = checkNode(check, context, undefined, options);
        return {
            ExpressionStatement: _checkNode
        };
    }
});

// The name of this rule.
const name$1 = "no-class";
// The default options for the rule.
const defaultOptions$1 = [];
// The possible error messages.
const errorMessages$1 = {
    generic: "Unexpected class, use functions not classes."
};
// The meta data for this rule.
const meta$1 = {
    type: "suggestion",
    docs: {
        description: "Disallow classes.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$1,
    schema: []
};
/**
 * Check if the given class node violates this rule.
 */
function checkClass(node, context) {
    // All class nodes violate this rule.
    context.report({ node, messageId: "generic" });
}
// Create the rule.
const rule$1 = createRule({
    name: name$1,
    meta: meta$1,
    defaultOptions: defaultOptions$1,
    create(context, options) {
        const _checkClass = checkNode(checkClass, context, undefined, options);
        return {
            ClassDeclaration: _checkClass,
            ClassExpression: _checkClass
        };
    }
});

// The name of this rule.
const name$2 = "no-delete";
// The default options for the rule.
const defaultOptions$2 = [];
// The possible error messages.
const errorMessages$2 = {
    generic: "Unexpected delete, objects should be considered immutable."
};
// The meta data for this rule.
const meta$2 = {
    type: "suggestion",
    docs: {
        description: "Disallow delete expressions.",
        category: "Best Practices",
        recommended: "error"
    },
    messages: errorMessages$2,
    schema: []
};
/**
 * Check if the given UnaryExpression violates this rule.
 */
function checkUnaryExpression(node, context) {
    if (node.operator === "delete") {
        context.report({ node, messageId: "generic" });
    }
}
// Create the rule.
const rule$2 = createRule({
    name: name$2,
    meta: meta$2,
    defaultOptions: defaultOptions$2,
    create(context, options) {
        const _checkUnaryExpression = checkNode(checkUnaryExpression, context, undefined, options);
        return {
            UnaryExpression: _checkUnaryExpression
        };
    }
});

// The name of this rule.
const name$3 = "no-expression-statement";
// The schema for the rule options.
const schema = [ignorePatternOptionsSchema];
// The default options for the rule.
const defaultOptions$3 = [{}];
// The possible error messages.
const errorMessages$3 = {
    generic: "Using expressions to cause side-effects not allowed."
};
// The meta data for this rule.
const meta$3 = {
    type: "suggestion",
    docs: {
        description: "Disallow expression statements.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$3,
    schema
};
/**
 * Check if the given ExpressionStatement violates this rule.
 */
function checkExpressionStatement(node, context) {
    // All expression statements violate this rule.
    context.report({ node, messageId: "generic" });
}
// Create the rule.
const rule$3 = createRule({
    name: name$3,
    meta: meta$3,
    defaultOptions: defaultOptions$3,
    create(context, [ignoreOptions, ...otherOptions]) {
        const _checkExpressionStatement = checkNode(checkExpressionStatement, context, ignoreOptions, otherOptions);
        return {
            ExpressionStatement: _checkExpressionStatement
        };
    }
});

// The name of this rule.
const name$4 = "no-if-statement";
// The default options for the rule.
const defaultOptions$4 = [];
// The possible error messages.
const errorMessages$4 = {
    generic: "Unexpected if, use a conditional expression (ternary operator) instead."
};
// The meta data for this rule.
const meta$4 = {
    type: "suggestion",
    docs: {
        description: "Disallow if statements.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$4,
    schema: []
};
/**
 * Check if the given IfStatement violates this rule.
 */
function checkIfStatement(node, context) {
    // All if statements violate this rule.
    context.report({ node, messageId: "generic" });
}
// Create the rule.
const rule$4 = createRule({
    name: name$4,
    meta: meta$4,
    defaultOptions: defaultOptions$4,
    create(context, options) {
        const _checkIfStatement = checkNode(checkIfStatement, context, undefined, options);
        return {
            IfStatement: _checkIfStatement
        };
    }
});

// The name of this rule.
const name$5 = "no-let";
// The schema for the rule options.
const schema$1 = [
    deepmerge.all([ignoreLocalOptionSchema, ignorePatternOptionsSchema])
];
// The default options for the rule.
const defaultOptions$5 = [
    {
        ignoreLocal: false
    }
];
// The possible error messages.
const errorMessages$5 = {
    generic: "Unexpected let, use const instead."
};
// The meta data for this rule.
const meta$5 = {
    type: "suggestion",
    docs: {
        description: "Disallow mutable variables.",
        category: "Best Practices",
        recommended: "error"
    },
    messages: errorMessages$5,
    fixable: "code",
    schema: schema$1
};
/**
 * Check if the given VariableDeclaration violates this rule.
 */
function checkVariableDeclaration(node, context) {
    if (node.kind === "let") {
        // Report the error.
        context.report({
            node,
            messageId: "generic",
            fix: 
            // Can only fix if all declarations have an initial value (with the
            // exception of ForOf and ForIn Statement initialisers).
            node.declarations.every(declaration => declaration.init !== null) ||
                isForXInitialiser(node)
                ? fixer => fixer.replaceTextRange([node.range[0], node.range[0] + node.kind.length], "const")
                : undefined
        });
    }
}
// Create the rule.
const rule$5 = createRule({
    name: name$5,
    meta: meta$5,
    defaultOptions: defaultOptions$5,
    create(context, [ignoreOptions, ...otherOptions]) {
        const _checkVariableDeclaration = checkNode(checkVariableDeclaration, context, ignoreOptions, otherOptions);
        return {
            VariableDeclaration: _checkVariableDeclaration
        };
    }
});

// The name of this rule.
const name$6 = "no-loop-statement";
// The default options for the rule.
const defaultOptions$6 = [];
// The possible error messages.
const errorMessages$6 = {
    generic: "Unexpected loop, use map or reduce instead."
};
// The meta data for this rule.
const meta$6 = {
    type: "suggestion",
    docs: {
        description: "Disallow imperative loops.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$6,
    schema: []
};
/**
 * Check if the given loop violates this rule.
 */
function checkLoop(node, context) {
    // All loops violate this rule.
    context.report({ node, messageId: "generic" });
}
// Create the rule.
const rule$6 = createRule({
    name: name$6,
    meta: meta$6,
    defaultOptions: defaultOptions$6,
    create(context, options) {
        const _checkLoop = checkNode(checkLoop, context, undefined, options);
        return {
            ForStatement: _checkLoop,
            ForInStatement: _checkLoop,
            ForOfStatement: _checkLoop,
            WhileStatement: _checkLoop,
            DoWhileStatement: _checkLoop
        };
    }
});

// The name of this rule.
const name$7 = "no-object-mutation";
// The default options for the rule.
const defaultOptions$7 = [];
// The possible error messages.
const errorMessages$7 = {
    generic: "Modifying properties of existing object not allowed."
};
// The meta data for this rule.
const meta$7 = {
    type: "suggestion",
    docs: {
        description: "Disallow mutating objects.",
        category: "Best Practices",
        recommended: "error"
    },
    messages: errorMessages$7,
    schema: []
};
/**
 * Check if the given node violates this rule.
 */
function check$1(node, context) {
    // TODO: port rule.
    context.report({ node, messageId: "generic" });
}
// Create the rule.
const rule$7 = createRule({
    name: name$7,
    meta: meta$7,
    defaultOptions: defaultOptions$7,
    create(context, options) {
        const _checkNode = checkNode(check$1, context, undefined, options);
        return {
            ExpressionStatement: _checkNode
        };
    }
});

// The name of this rule.
const name$8 = "no-reject";
// The default options for the rule.
const defaultOptions$8 = [];
// The possible error messages.
const errorMessages$8 = {
    generic: "Unexpected reject, return an error instead."
};
// The meta data for this rule.
const meta$8 = {
    type: "suggestion",
    docs: {
        description: "Disallow try-catch[-finally] and try-finally patterns.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$8,
    schema: []
};
/**
 * Check if the given CallExpression violates this rule.
 */
function checkCallExpression(node, context) {
    if (isMemberExpression(node.callee) &&
        isIdentifier(node.callee.object) &&
        isIdentifier(node.callee.property) &&
        node.callee.object.name === "Promise" &&
        node.callee.property.name === "reject") {
        context.report({ node, messageId: "generic" });
    }
}
// Create the rule.
const rule$8 = createRule({
    name: name$8,
    meta: meta$8,
    defaultOptions: defaultOptions$8,
    create(context, options) {
        const _checkCallExpression = checkNode(checkCallExpression, context, undefined, options);
        return {
            CallExpression: _checkCallExpression
        };
    }
});

// The name of this rule.
const name$9 = "no-this";
// The default options for the rule.
const defaultOptions$9 = [];
// The possible error messages.
const errorMessages$9 = {
    generic: "Unexpected this, use functions not classes."
};
// The meta data for this rule.
const meta$9 = {
    type: "suggestion",
    docs: {
        description: "Disallow this access.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$9,
    schema: []
};
/**
 * Check if the given ThisExpression violates this rule.
 */
function checkThisExpression(node, context) {
    // All throw statements violate this rule.
    context.report({ node, messageId: "generic" });
}
// Create the rule.
const rule$9 = createRule({
    name: name$9,
    meta: meta$9,
    defaultOptions: defaultOptions$9,
    create(context, options) {
        const _checkThisExpression = checkNode(checkThisExpression, context, undefined, options);
        return {
            ThisExpression: _checkThisExpression
        };
    }
});

// The name of this rule.
const name$a = "no-throw";
// The default options for the rule.
const defaultOptions$a = [];
// The possible error messages.
const errorMessages$a = {
    generic: "Unexpected throw, throwing exceptions is not functional."
};
// The meta data for this rule.
const meta$a = {
    type: "suggestion",
    docs: {
        description: "Disallow throwing exceptions.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$a,
    schema: []
};
/**
 * Check if the given ThrowStatement violates this rule.
 */
function checkThrowStatement(node, context) {
    // All throw statements violate this rule.
    context.report({ node, messageId: "generic" });
}
// Create the rule.
const rule$a = createRule({
    name: name$a,
    meta: meta$a,
    defaultOptions: defaultOptions$a,
    create(context, options) {
        const _checkThrowStatement = checkNode(checkThrowStatement, context, undefined, options);
        return {
            ThrowStatement: _checkThrowStatement
        };
    }
});

// The name of this rule.
const name$b = "no-try";
// The default options for the rule.
const defaultOptions$b = [];
// The possible error messages.
const errorMessages$b = {
    generic: "Unexpected try, the try-catch[-finally] and try-finally patterns are not functional."
};
// The meta data for this rule.
const meta$b = {
    type: "suggestion",
    docs: {
        description: "Disallow try-catch[-finally] and try-finally patterns.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$b,
    schema: []
};
/**
 * Check if the given TryStatement violates this rule.
 */
function checkTryStatement(node, context) {
    // All try statements violate this rule.
    context.report({ node, messageId: "generic" });
}
// Create the rule.
const rule$b = createRule({
    name: name$b,
    meta: meta$b,
    defaultOptions: defaultOptions$b,
    create(context, options) {
        const _checkTryStatement = checkNode(checkTryStatement, context, undefined, options);
        return {
            TryStatement: _checkTryStatement
        };
    }
});

// The name of this rule.
const name$c = "no-method-signature";
// The default options for the rule.
const defaultOptions$c = [];
// The possible error messages.
const errorMessages$c = {
    generic: "Method signature is mutable, use property signature with readonly modifier instead."
};
// The meta data for this rule.
const meta$c = {
    type: "suggestion",
    docs: {
        description: "Prefer property signatures with readonly modifiers over method signatures.",
        category: "Best Practices",
        recommended: "error"
    },
    messages: errorMessages$c,
    schema: []
};
/**
 * Check if the given TSMethodSignature violates this rule.
 */
function checkTSMethodSignature(node, context) {
    // All TS method signatures violate this rule.
    context.report({ node, messageId: "generic" });
}
// Create the rule.
const rule$c = createRule({
    name: name$c,
    meta: meta$c,
    defaultOptions: defaultOptions$c,
    create(context, options) {
        const _checkTSMethodSignature = checkNode(checkTSMethodSignature, context, undefined, options);
        return {
            TSMethodSignature: _checkTSMethodSignature
        };
    }
});

// The name of this rule.
const name$d = "no-mixed-interface";
// The default options for the rule.
const defaultOptions$d = [];
// The possible error messages.
const errorMessages$d = {
    generic: "Only the same kind of members allowed in interfaces."
};
// The meta data for this rule.
const meta$d = {
    type: "suggestion",
    docs: {
        description: "Restrict interfaces so that only members of the same kind of are allowed in them.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$d,
    schema: []
};
/**
 * Check if the given TSInterfaceDeclaration violates this rule.
 */
function checkTSInterfaceDeclaration(node, context) {
    let prevMemberType = undefined;
    let prevMemberTypeAnnotation = undefined;
    for (const member of node.body.body) {
        const memberType = member.type;
        const memberTypeAnnotation = isTSPropertySignature(member) && member.typeAnnotation !== undefined
            ? member.typeAnnotation.typeAnnotation.type
            : undefined;
        if (
        // Not the first property in the interface.
        prevMemberType !== undefined &&
            // And different property type to previous property.
            (prevMemberType !== memberType ||
                // Or annotationed with a different type annotation.
                (prevMemberTypeAnnotation !== memberTypeAnnotation &&
                    // Where one of the properties is a annotationed as a function.
                    (prevMemberTypeAnnotation === astNodeTypes.AST_NODE_TYPES.TSFunctionType ||
                        memberTypeAnnotation === astNodeTypes.AST_NODE_TYPES.TSFunctionType)))) {
            context.report({ node: member, messageId: "generic" });
        }
        prevMemberType = memberType;
        prevMemberTypeAnnotation = memberTypeAnnotation;
    }
}
// Create the rule.
const rule$d = createRule({
    name: name$d,
    meta: meta$d,
    defaultOptions: defaultOptions$d,
    create(context, options) {
        const _checkTSInterfaceDeclaration = checkNode(checkTSInterfaceDeclaration, context, undefined, options);
        return {
            TSInterfaceDeclaration: _checkTSInterfaceDeclaration
        };
    }
});

// The name of this rule.
const name$e = "readonly-array";
// The schema for the rule options.
const schema$2 = [
    deepmerge.all([
        ignoreLocalOptionSchema,
        ignorePatternOptionsSchema,
        ignoreReturnTypeOptionSchema
    ])
];
// The default options for the rule.
const defaultOptions$e = [
    {
        ignoreLocal: false,
        ignoreReturnType: false
    }
];
// The possible error messages.
const errorMessages$e = {
    generic: "Only ReadonlyArray allowed."
};
// The meta data for this rule.
const meta$e = {
    type: "suggestion",
    docs: {
        description: "Prefer readonly array over mutable arrays.",
        category: "Best Practices",
        recommended: "error"
    },
    messages: errorMessages$e,
    fixable: "code",
    schema: schema$2
};
/**
 * Check if the given ArrayType or TupleType violates this rule.
 */
function checkArrayOrTupleType(node, context, [options]) {
    if (!node.parent ||
        !isTSTypeOperator(node.parent) ||
        node.parent.operator !== "readonly") {
        if (options.ignoreReturnType && isInReturnType(node)) {
            return;
        }
        context.report({
            node,
            messageId: "generic",
            fix: fixer => node.parent && isTSArrayType(node.parent)
                ? [
                    fixer.insertTextBefore(node, "(readonly "),
                    fixer.insertTextAfter(node, ")")
                ]
                : fixer.insertTextBefore(node, "readonly ")
        });
    }
}
/**
 * Check if the given TypeReference violates this rule.
 */
function checkTypeReference(node, context, [options]) {
    if (isIdentifier(node.typeName) && node.typeName.name === "Array") {
        if (options.ignoreReturnType && isInReturnType(node)) {
            return;
        }
        context.report({
            node,
            messageId: "generic",
            fix: fixer => fixer.insertTextBefore(node, "Readonly")
        });
    }
}
/**
 * Is the given node in the return type.
 */
function isInReturnType(node) {
    let n = node;
    while (n && n.parent) {
        if (isFunctionLike(n.parent)) {
            if (n.parent.returnType === n) {
                return true;
            }
            return false;
        }
        n = n.parent;
    }
    return false;
}
// Create the rule.
const rule$e = createRule({
    name: name$e,
    meta: meta$e,
    defaultOptions: defaultOptions$e,
    create(context, options) {
        const _checkArrayOrTupleType = checkNode(checkArrayOrTupleType, context, undefined, options);
        const _checkTypeReference = checkNode(checkTypeReference, context, undefined, options);
        return {
            TSArrayType: _checkArrayOrTupleType,
            TSTupleType: _checkArrayOrTupleType,
            TSTypeReference: _checkTypeReference
        };
    }
});

// The name of this rule.
const name$f = "readonly-keyword";
// The schema for the rule options.
const schema$3 = [
    deepmerge.all([
        ignoreLocalOptionSchema,
        ignorePatternOptionsSchema,
        ignoreClassOptionSchema,
        ignoreInterfaceOptionSchema
    ])
];
// The default options for the rule.
const defaultOptions$f = [
    {
        ignoreClass: false,
        ignoreInterface: false,
        ignoreLocal: false
    }
];
// The possible error messages.
const errorMessages$f = {
    generic: "A readonly modifier is required."
};
// The meta data for this rule.
const meta$f = {
    type: "suggestion",
    docs: {
        description: "Enforce readonly modifiers are used where possible.",
        category: "Best Practices",
        recommended: "error"
    },
    messages: errorMessages$f,
    fixable: "code",
    schema: schema$3
};
/**
 * Check if the given node violates this rule.
 */
function check$2(node, context) {
    if (!node.readonly) {
        const fix = "readonly ";
        context.report({
            node,
            messageId: "generic",
            fix: isTSIndexSignature(node)
                ? fixer => fixer.insertTextBefore(node, fix)
                : fixer => fixer.insertTextBefore(node.key, fix)
        });
    }
}
// Create the rule.
const rule$f = createRule({
    name: name$f,
    meta: meta$f,
    defaultOptions: defaultOptions$f,
    create(context, [ignoreOptions, ...otherOptions]) {
        const _checkNode = checkNode(check$2, context, ignoreOptions, otherOptions);
        return {
            ClassProperty: _checkNode,
            TSIndexSignature: _checkNode,
            TSPropertySignature: _checkNode
        };
    }
});

/**
 * All of the custom rules.
 */
const rules = {
    [name]: rule,
    [name$1]: rule$1,
    [name$2]: rule$2,
    [name$3]: rule$3,
    [name$4]: rule$4,
    [name$5]: rule$5,
    [name$6]: rule$6,
    [name$c]: rule$c,
    [name$d]: rule$d,
    [name$7]: rule$7,
    [name$8]: rule$8,
    [name$9]: rule$9,
    [name$a]: rule$a,
    [name$b]: rule$b,
    [name$e]: rule$e,
    [name$f]: rule$f
};

const config = {
    rules,
    configs: {
        all,
        base,
        recommended,
        functional,
        "functional-lite": functionalLite,
        "eslint-recommended": eslintRecommended
    }
};

module.exports = config;
